SystemOrganization addCategory: #'SmaCC-Runtime'!

Error subclass: #SmaCCParserError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

!SmaCCParserError commentStamp: 'jmb' prior: 0!
SmaCCParserException is the exception raised when a parsing error occurs. The description of the exception will be the error message and the parameter of the exception is the parser. With this information, you can insert a custom error message in your text view that you are parsing. For example, in VisualWorks, the following code will insert an error message into your text view:
	textController insertAndSelect: ex description , ' ->' at: ex parameter position!

Object subclass: #SmaCCParser
	instanceVariableNames: 'scanner currentToken errorToken stateStack nodeStack'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

!SmaCCParser commentStamp: 'jmb' prior: 0!
SmaCCParser is an abstract class that defines most of the parsing actions. Subclasses will define methods that specify their transitions and reduction actions. These are normally defined automatically when compiling the parser.

Subclasses must implement the following messages:
	accessing
		emptySymbolTokenId
		reduceTable
		transitionTable

Instance Variables:
	currentToken	<SmaCCToken>	the token last returned by the scanner that has not been shifted (reduce actions leave the current token alone)
	nodeStack	<OrderedCollection>	collection of items on stack. These items are specific to the parser and can be any object. 
	scanner	<SmaCCScanner>	our scanner
	stateStack	<OrderedCollection of: Integer>	the stack of states for our parser (standard LR state stack)

!

!SmaCCParser classSide methodsFor: 'private' stamp: ''!
defaultStartingState
	^1! !

!SmaCCParser classSide methodsFor: 'instance creation' stamp: ''!
on: aStream 
	| parser scanner |
	scanner := self scannerClass on: aStream.
	parser := self new.
	parser scanner: scanner.
	^parser! !

!SmaCCParser classSide methodsFor: 'accessing' stamp: ''!
parse: aString 
	^self parse: aString startingAt: self defaultStartingState! !

!SmaCCParser classSide methodsFor: 'accessing' stamp: 'jmb 1/20/2003 22:09'!
parse: aString onError: aBlock 
	^[self parse: aString] on: SmaCCParserError
		do: [:ex | ex return: (aBlock value: ex description value: ex tag position)]! !

!SmaCCParser classSide methodsFor: 'accessing' stamp: 'damiencassou 5/30/2008 16:02'!
parse: aString startingAt: anInteger 
	^self parseStream: aString readStream startingAt: anInteger! !

!SmaCCParser classSide methodsFor: 'accessing' stamp: 'jmb 1/20/2003 22:04'!
parse: aString startingAt: anInteger onError: aBlock 
	^[self parse: aString startingAt: anInteger] on: SmaCCParserError
		do: [:ex | ex return: (aBlock value: ex description value: ex tag position)]! !

!SmaCCParser classSide methodsFor: 'accessing' stamp: ''!
parseStream: aStream 
	^self parseStream: aStream startingAt: self defaultStartingState! !

!SmaCCParser classSide methodsFor: 'accessing' stamp: 'jmb 1/20/2003 22:11'!
parseStream: aStream onError: aBlock 
	^[self parseStream: aStream] on: SmaCCParserError
		do: [:ex | ex return: (aBlock value: ex description value: ex tag position)]! !

!SmaCCParser classSide methodsFor: 'accessing' stamp: ''!
parseStream: aStream startingAt: anInteger 
	| parser |
	parser := self on: aStream.
	parser setStartingState: anInteger.
	^parser parse! !

!SmaCCParser classSide methodsFor: 'accessing' stamp: 'md 9/4/2005 16:58'!
parseStream: aStream startingAt: anInteger onError: aBlock 
	^[self parseStream: aStream startingAt: anInteger] 
		on: SmaCCParserError
		do: [:ex | ex return: (aBlock value: ex description value: ex tag position)]! !

!SmaCCParser classSide methodsFor: 'private' stamp: ''!
scannerClass
	^self subclassResponsibility! !

!SmaCCParser methodsFor: 'private' stamp: ''!
acceptAction
	^0! !

!SmaCCParser methodsFor: 'private' stamp: ''!
actionFor: aSymbolIndex 
	^self actionForState: self currentState and: aSymbolIndex! !

!SmaCCParser methodsFor: 'private' stamp: ''!
actionForCurrentToken
	^self actionFor: currentToken id first! !

!SmaCCParser methodsFor: 'private' stamp: 'fbs 2/13/2006 23:08'!
actionForState: stateIndex and: aSymbolIndex 
	| index row |
	row := self transitionTable at: stateIndex.
	(row at: 1) = 2 
		ifTrue: 
			[index := self 
						binarySearchIn: row
						for: aSymbolIndex
						size: 1.
			index = 0 ifTrue: [^self errorAction] ifFalse: [^row at: 2]]
		ifFalse: 
			[index := self 
						binarySearchIn: row
						for: aSymbolIndex
						size: 2.
			index = 0 ifTrue: [^self errorAction] ifFalse: [^row at: index - 1]]! !

!SmaCCParser methodsFor: 'private' stamp: ''!
actionMask
	^2r11! !

!SmaCCParser methodsFor: 'private' stamp: ''!
binarySearchIn: aRow for: aSymbolIndex size: step 
	| start mid length midItem stop |
	start := 3.
	stop := aRow size.
	length := (stop - start) // step.
	[length > 4] whileTrue: 
			[length := length bitShift: -1.
			mid := length * step + start.
			midItem := aRow at: mid.
			midItem <= aSymbolIndex ifTrue: [start := mid] ifFalse: [stop := mid]].
	[start <= stop] whileTrue: 
			[(aRow at: start) == aSymbolIndex ifTrue: [^start].
			start := start + step].
	^0! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ''!
checkForErrors
	"If we have an error correction installed, we might have handled the errors. If we did, we don't 
	want to return the result, so we raise a final exception that can't be proceeded."

	errorToken isNil ifTrue: [^self].
	currentToken := errorToken.
	self reportErrorMessage: 'Token not expected'! !

!SmaCCParser methodsFor: 'private' stamp: ''!
currentState
	^stateStack last! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ''!
dismissErrorToken
	currentToken := nil.
	self getNextToken! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ''!
dismissStackTopForErrorRecovery
	stateStack removeLast.
	^nodeStack removeLast! !

!SmaCCParser methodsFor: 'accessing' stamp: ''!
emptySymbolTokenId
	^scanner emptySymbolTokenId! !

!SmaCCParser methodsFor: 'private' stamp: ''!
errorAction
	^3! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ''!
errorHandlerStates
	^stateStack collect: 
			[:each | 
			| action |
			action := self actionForState: each and: self errorTokenId.
			(action bitAnd: self actionMask) = 1 
				ifTrue: [action bitShift: -2]
				ifFalse: [0]]! !

!SmaCCParser methodsFor: 'accessing' stamp: ''!
errorTable
	^#()! !

!SmaCCParser methodsFor: 'accessing' stamp: ''!
errorTokenId
	^scanner errorTokenId! !

!SmaCCParser methodsFor: 'private' stamp: ''!
findErrorHandlerIfNoneUseErrorNumber: anInteger 
	| handlerStates index startingErrorToken newStack |
	handlerStates := self errorHandlerStates reverse.
	startingErrorToken := currentToken.
	
	[index := (1 to: handlerStates size) detect: 
					[:each | 
					| state |
					state := handlerStates at: each.
					state ~= 0 and: 
							[newStack := stateStack copyFrom: 1 to: handlerStates size - each + 1.
							newStack add: state.
							self willShift: newStack]]
				ifNone: [nil].
	index isNil] 
			whileTrue: 
				[self dismissErrorToken.
				currentToken id first = self emptySymbolTokenId 
					ifTrue: 
						[currentToken := startingErrorToken.
						self reportError: anInteger]].
	index - 1 timesRepeat: [self dismissStackTopForErrorRecovery].
	stateStack addLast: (handlerStates at: index).
	nodeStack addLast: startingErrorToken! !

!SmaCCParser methodsFor: 'private' stamp: ''!
getNextToken
	currentToken isNil ifTrue: [currentToken := scanner next]! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ''!
handleError: anInteger 
	errorToken isNil ifTrue: [errorToken := currentToken].
	(currentToken id first = self emptySymbolTokenId 
		or: [self hasErrorHandler not]) ifTrue: [self reportError: anInteger].
	self findErrorHandlerIfNoneUseErrorNumber: anInteger! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ''!
hasErrorHandler
	^self errorHandlerStates anySatisfy: [:each | each ~~ 0]! !

!SmaCCParser methodsFor: 'initialization' stamp: ''!
initialize
	nodeStack := OrderedCollection new! !

!SmaCCParser methodsFor: 'testing' stamp: ''!
isEOFToken
	^currentToken id first = self emptySymbolTokenId! !

!SmaCCParser methodsFor: 'private' stamp: ''!
liftFirstValue: aCollection 
	^aCollection first! !

!SmaCCParser methodsFor: 'private' stamp: ''!
liftLastValue: aCollection 
	^aCollection last! !

!SmaCCParser methodsFor: 'private' stamp: ''!
liftSecondValue: aCollection 
	^aCollection at: 2! !

!SmaCCParser methodsFor: 'accessing' stamp: ''!
parse
	self setDefaultStartingStateIfNone.
	self performParsingLoop.
	^nodeStack last! !

!SmaCCParser methodsFor: 'private' stamp: ''!
performParsingLoop
	| action actionType |
	
	[self getNextToken.
	action := self actionForCurrentToken.
	action = self acceptAction] 
			whileFalse: 
				[actionType := action bitAnd: self actionMask.
				action := action bitShift: -2.
				actionType == self shiftAction 
					ifTrue: [self shift: action]
					ifFalse: 
						[actionType == self reduceAction 
							ifTrue: [self reduce: action]
							ifFalse: [self handleError: action]]].
	self checkForErrors! !

!SmaCCParser methodsFor: 'private' stamp: ''!
performReduceMethod: aSymbol with: items 
	^aSymbol last == $: 
		ifTrue: [self perform: aSymbol with: items]
		ifFalse: [self perform: aSymbol]! !

!SmaCCParser methodsFor: 'accessing' stamp: ''!
position
	^currentToken isNil 
		ifTrue: [scanner position]
		ifFalse: [currentToken startPosition]! !

!SmaCCParser methodsFor: 'private' stamp: ''!
reduce: anInteger 
	| reduceEntry items size |
	reduceEntry := self reduceTable at: anInteger.
	items := OrderedCollection new: (size := reduceEntry at: 2).
	size timesRepeat: 
			[items addFirst: nodeStack removeLast.
			stateStack removeLast].
	nodeStack add: (self performReduceMethod: (reduceEntry at: 3) with: items).
	stateStack add: ((self actionFor: (reduceEntry at: 1)) bitShift: -2)! !

!SmaCCParser methodsFor: 'private' stamp: ''!
reduceAction
	^2r10! !

!SmaCCParser methodsFor: 'private' stamp: ''!
reduceFor: aCollection 
	| newCollection item |
	(aCollection allSatisfy: [:each | each class ~~ OrderedCollection]) 
		ifTrue: [^aCollection].
	aCollection first class == OrderedCollection 
		ifTrue: 
			[newCollection := aCollection first.
			2 to: aCollection size
				do: 
					[:i | 
					item := aCollection at: i.
					item class = OrderedCollection 
						ifTrue: [newCollection addAll: item]
						ifFalse: [newCollection add: item]].
			^newCollection].
	newCollection := OrderedCollection new.
	aCollection do: 
			[:each | 
			each class == OrderedCollection 
				ifTrue: [newCollection addAll: each]
				ifFalse: [newCollection add: each]].
	^newCollection! !

!SmaCCParser methodsFor: 'accessing' stamp: ''!
reduceTable
	^self subclassResponsibility! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ''!
reportError: anInteger 
	self reportErrorMessage: (anInteger = 0 
				ifTrue: ['Token not expected']
				ifFalse: [self errorTable at: anInteger])! !

!SmaCCParser methodsFor: 'private-error handling' stamp: 'jmb 1/20/2003 21:52'!
reportErrorMessage: aString 
	(SmaCCParserError new)
		tag: self;
		signal: aString! !

!SmaCCParser methodsFor: 'initialization' stamp: ''!
scanner: aScanner 
	scanner := aScanner! !

!SmaCCParser methodsFor: 'private' stamp: ''!
setDefaultStartingStateIfNone
	stateStack isNil 
		ifTrue: [self setStartingState: self class defaultStartingState]! !

!SmaCCParser methodsFor: 'initialization' stamp: ''!
setStartingState: startingState 
	stateStack := OrderedCollection with: startingState! !

!SmaCCParser methodsFor: 'private' stamp: ''!
shift: stateIndex 
	stateStack add: stateIndex.
	nodeStack add: currentToken.
	currentToken := nil! !

!SmaCCParser methodsFor: 'private' stamp: ''!
shiftAction
	^2r01! !

!SmaCCParser methodsFor: 'accessing' stamp: 'md 11/9/2004 11:11'!
stringValue: anOrderedCollection
	^anOrderedCollection first value.! !

!SmaCCParser methodsFor: 'accessing' stamp: ''!
transitionTable
	^self subclassResponsibility! !

!SmaCCParser methodsFor: 'private-error handling' stamp: ''!
willShift: potentialStateStack 
	| action compoundAction reduceEntry size |
	compoundAction := self actionForState: potentialStateStack last
				and: currentToken id first.
	action := compoundAction bitAnd: self actionMask.
	action == self shiftAction ifTrue: [^true].
	action == self reduceAction 
		ifTrue: 
			[reduceEntry := self reduceTable at: (compoundAction bitShift: -2).
			size := reduceEntry at: 2.
			size timesRepeat: [potentialStateStack removeLast].
			potentialStateStack 
				add: ((self actionForState: potentialStateStack last
						and: (reduceEntry at: 1)) bitShift: -2).
			^self willShift: potentialStateStack].
	^false! !

Object subclass: #SmaCCScanner
	instanceVariableNames: 'stream start matchActions matchEnd currentCharacter outputStream lastOutputStreamMatchPosition lastMatchWasEmpty returnMatchBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

SmaCCScanner class
	instanceVariableNames: 'keywordMap'!

!SmaCCScanner commentStamp: 'jmb' prior: 0!
SmaCCScanner is an abstract class that represents a scanner for the parser. The scanner converts its string input into SmaCCToken objects that the parser then uses for its parsing.

Subclasses must implement the following messages:
	accessing
		scanForToken

Instance Variables:
	currentCharacter	<Character>	the current character we are scanning
	lastMatchWasEmpty	<Boolean>	was our last scanning match an empty string -- don't allow two empty matches in a row
	lastOutputStreamMatchPosition	<Integer>	the position in the outputStream of the last match
	matchActions	<Array | Symbol>	the actions for the last match (a symbol means that the action should be performed on the scanner)
	matchEnd	<Integer>	the position of the last match in the stream (our input stream)
	outputStream	<PositionableStream>	the matched characters go in this stream. After a match is made, we take this stream's contents and create a token object.
	returnMatchBlock	<BlockClosure>	when we match a token evaluate this block with the token (hack to return from multiple levels)
	start	<Integer>	the starting position of a match in the stream
	stream	<Stream>	our input

!

SmaCCScanner class
	instanceVariableNames: 'keywordMap'!

!SmaCCScanner classSide methodsFor: 'accessing' stamp: 'mga 2/8/2004 22:35'!
frequencyTable
	^#(1)! !

!SmaCCScanner classSide methodsFor: 'initialization' stamp: ''!
initialize
	self initializeKeywordMap! !

!SmaCCScanner classSide methodsFor: 'initialization' stamp: ''!
initializeKeywordMap
	keywordMap := Dictionary new! !

!SmaCCScanner classSide methodsFor: 'accessing' stamp: ''!
keywordMap
	keywordMap isNil ifTrue: [self initializeKeywordMap].
	^keywordMap! !

!SmaCCScanner classSide methodsFor: 'testing' stamp: ''!
needsLineNumbers
	"Redefine to return true, if you need line number information"

	^false! !

!SmaCCScanner classSide methodsFor: 'instance creation' stamp: ''!
on: aStream 
	^(self new)
		on: (self needsLineNumbers 
					ifTrue: [SmaCCLineNumberStream on: aStream]
					ifFalse: [aStream]);
		yourself! !

!SmaCCScanner methodsFor: 'testing' stamp: ''!
atEnd
	^stream atEnd! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
checkForKeyword: aString 
	| stateMap action |
	action := matchActions isSymbol 
				ifTrue: [matchActions]
				ifFalse: [matchActions first].
	stateMap := self class keywordMap at: action ifAbsent: [nil].
	stateMap isNil ifTrue: [^self].
	matchActions := stateMap at: (self keywordFor: aString)
				ifAbsent: [matchActions].
	matchActions isInteger 
		ifTrue: [matchActions := Array with: matchActions with: action]! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
checkForValidMatch
	matchActions isNil ifTrue: [self scannerError]! !

!SmaCCScanner methodsFor: 'default token handling' stamp: ''!
comment
	"In case someone wants to record the comments"

	self whitespace! !

!SmaCCScanner methodsFor: 'accessing' stamp: ''!
contents
	| writeStream token |
	writeStream := WriteStream on: Array new.
	[self atEnd] whileFalse: 
			[token := self next.
			token notNil ifTrue: [writeStream nextPut: token]].
	^writeStream contents! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
createTokenFor: string 
	| token |
	token := SmaCCToken 
				value: string
				start: start
				id: matchActions.
	outputStream reset.
	matchActions := nil.
	returnMatchBlock value: token! !

!SmaCCScanner methodsFor: 'accessing' stamp: ''!
emptySymbolTokenId
	^self subclassResponsibility! !

!SmaCCScanner methodsFor: 'accessing' stamp: ''!
errorTokenId
	^self subclassResponsibility! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
initialBufferSize
	^128! !

!SmaCCScanner methodsFor: 'initialization' stamp: ''!
initialize
	outputStream := WriteStream on: (String new: self initialBufferSize).
	lastMatchWasEmpty := true! !

!SmaCCScanner methodsFor: 'private-utility' stamp: ''!
keywordFor: aString 
	"Subclasses can override this to ignore case"

	^aString! !

!SmaCCScanner methodsFor: 'accessing' stamp: ''!
lineNumber
	"This requires the stream to be a line number stream (see the #needsLineNumbers class method)."

	^stream lineNumber! !

!SmaCCScanner methodsFor: 'accessing' stamp: ''!
next
	self resetScanner.
	returnMatchBlock := [:match | ^match].
	self scanForToken! !

!SmaCCScanner methodsFor: 'initialization' stamp: 'ajh 3/17/2003 12:05'!
on: aStream 
	stream := aStream.
	start := stream position + 1.! !

!SmaCCScanner methodsFor: 'accessing' stamp: ''!
position
	^stream position! !

!SmaCCScanner methodsFor: 'accessing' stamp: ''!
position: anInteger
	^stream position: anInteger! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
recordAndReportMatch: aCollection 
	self
		recordMatch: aCollection;
		reportLastMatch! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
recordMatch: aCollection 
	matchActions := aCollection.
	matchEnd := stream position.
	lastOutputStreamMatchPosition := outputStream position! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
reportLastMatch
	"The scanner has found the end of a token and must report it"

	| string |
	self checkForValidMatch.
	self resetOutputToLastMatch.
	stream position: matchEnd.
	string := outputStream contents.
	self checkForKeyword: string.
	matchActions isSymbol 
		ifTrue: [self perform: matchActions]
		ifFalse: [self createTokenFor: string]! !

!SmaCCScanner methodsFor: 'private' stamp: 'fbs 2/13/2006 22:31'!
resetOutputToLastMatch
	outputStream position: lastOutputStreamMatchPosition.
	lastOutputStreamMatchPosition = 0 
		ifTrue: 
			[lastMatchWasEmpty ifTrue: [self scannerError].
			lastMatchWasEmpty := true]
		ifFalse: [lastMatchWasEmpty := false]! !

!SmaCCScanner methodsFor: 'private' stamp: 'ajh 3/17/2003 12:04'!
resetScanner
	start := stream position + 1.
	outputStream reset.
	lastOutputStreamMatchPosition := 0! !

!SmaCCScanner methodsFor: 'accessing' stamp: ''!
scanForToken
	^self subclassResponsibility! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
scannerError
	(stream atEnd and: [start == (stream position + 1)]) 
		ifTrue: 
			[returnMatchBlock value: (SmaCCToken 
						value: ''
						start: start
						id: (Array with: self emptySymbolTokenId))].
	stream position: start - 1.
	returnMatchBlock value: (SmaCCToken 
				value: (String with: stream next)
				start: start
				id: #(0))! !

!SmaCCScanner methodsFor: 'private' stamp: ''!
step
	stream atEnd ifTrue: [^self reportLastMatch].
	currentCharacter := stream next.
	outputStream nextPut: currentCharacter! !

!SmaCCScanner methodsFor: 'default token handling' stamp: ''!
whitespace
	"By default, eat the whitespace"

	self resetScanner.
	self scanForToken! !

Object subclass: #SmaCCToken
	instanceVariableNames: 'start id value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

!SmaCCToken commentStamp: 'jmb' prior: 0!
SmaCCTokens are used as the interface objects between scanner and parser. They hold the string that was scanned and its position information. Also, included in the token is its id. The id specifies what type of token it is.

Instance Variables:
	id	<Array of: Integer>	the list of possible token types this represents. There can be overlapping tokens, so we list all of the id here. The default parser only looks at the first id, but we can redefine this behavior in a subclass to look at all possibilities until we find a valid token.
	start	<Integer>	the starting position of the token in the original input
	value	<Object>	the value of our token (normally a string, but could be anything)
!

!SmaCCToken classSide methodsFor: 'instance creation' stamp: 'md 7/12/2006 19:29'!
value: aString

	^ self value: aString start: 0 id: nil! !

!SmaCCToken classSide methodsFor: 'instance creation' stamp: 'ajh 3/11/2003 23:28'!
value: aString start: anInteger

	^ self value: aString start: anInteger id: nil! !

!SmaCCToken classSide methodsFor: 'instance creation' stamp: ''!
value: aString start: anInteger id: anObject 
	^(self new)
		value: aString
			start: anInteger
			id: anObject;
		yourself! !

!SmaCCToken methodsFor: 'accessing' stamp: ''!
id
	^id! !

!SmaCCToken methodsFor: 'testing' stamp: 'md 9/6/2005 17:22'!
isRBToken
	^false! !

!SmaCCToken methodsFor: 'accessing' stamp: 'ajh 2/27/2003 17:55'!
length

	^ value size! !

!SmaCCToken methodsFor: 'printing' stamp: ''!
printOn: aStream 
	aStream
		nextPut: ${;
		nextPutAll: self value;
		nextPut: $(;
		nextPutAll: self startPosition printString;
		nextPut: $,;
		nextPutAll: self stopPosition printString;
		nextPut: $,;
		nextPutAll: self id printString;
		nextPutAll: ')}'! !

!SmaCCToken methodsFor: 'accessing' stamp: 'md 8/12/2005 15:09'!
realValue
	^value! !

!SmaCCToken methodsFor: 'accessing' stamp: 'ajh 3/11/2003 23:05'!
sourceInterval

	^ self start to: self stop! !

!SmaCCToken methodsFor: 'accessing' stamp: 'ajh 3/17/2003 12:03'!
start

	^ start! !

!SmaCCToken methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:36'!
start: startPositionInteger

	start := startPositionInteger
! !

!SmaCCToken methodsFor: 'accessing' stamp: 'ajh 3/17/2003 12:03'!
startPosition
	^ start! !

!SmaCCToken methodsFor: 'accessing' stamp: 'ajh 3/17/2003 12:03'!
stop

	^ start ifNotNil: [start + value size - 1]! !

!SmaCCToken methodsFor: 'accessing' stamp: 'md 7/12/2006 19:41'!
stopPosition
	^ start ifNotNil: [start + value size - 1].! !

!SmaCCToken methodsFor: 'accessing' stamp: ''!
value
	^value! !

!SmaCCToken methodsFor: 'initialization' stamp: 'stephaneducasse 2/4/2006 20:36'!
value: anObject

	value := anObject! !

!SmaCCToken methodsFor: 'initialization' stamp: ''!
value: anObject start: startPositionInteger id: anInteger 
	value := anObject.
	start := startPositionInteger.
	id := anInteger! !

Stream subclass: #SmaCCLineNumberStream
	instanceVariableNames: 'sourceStream previousWasCR eolPositions lastPosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'SmaCC-Runtime'!

!SmaCCLineNumberStream commentStamp: 'jmb' prior: 0!
SmaCCLineNumberStream is a wrapper for streams that calculates line numbers.

Instance Variables:
	eolPositions	<OrderedCollection of: Integer>	the positions of each end of line
	lastPosition	<Integer>	the position of the last character that we have calculated the end of line information for (we know the line number for all characters before this position and don't know anything about the characters after this position)
	previousWasCR	<Boolean>	was the previous character a CR. This is used for CR LF streams. A CR LF combination should only increment the line counter by 1
	sourceStream	<Stream>	the stream that we are wrapping

!

!SmaCCLineNumberStream classSide methodsFor: 'instance creation' stamp: ''!
on: aReadStream 
	^(self basicNew)
		on: aReadStream;
		yourself! !

!SmaCCLineNumberStream methodsFor: 'testing' stamp: ''!
atEnd
	^sourceStream atEnd! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ''!
contents
	^sourceStream contents! !

!SmaCCLineNumberStream methodsFor: 'error handling' stamp: ''!
doesNotUnderstand: aMessage 
	^sourceStream perform: aMessage selector withArguments: aMessage arguments! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ''!
flush
	^sourceStream flush! !

!SmaCCLineNumberStream methodsFor: 'testing' stamp: ''!
isReadable
	^sourceStream isReadable! !

!SmaCCLineNumberStream methodsFor: 'testing' stamp: ''!
isWritable
	^sourceStream isWritable! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ''!
lineNumber
	| index start stop pos |
	pos := sourceStream position.
	pos >= eolPositions last ifTrue: [^eolPositions size].
	start := 1.
	stop := eolPositions size.
	[start + 1 < stop] whileTrue: 
			[index := (start + stop) // 2.
			(eolPositions at: index) <= pos 
				ifTrue: [start := index]
				ifFalse: [stop := index]].
	^start! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ''!
next
	| character |
	character := sourceStream next.
	sourceStream position - 1 == lastPosition 
		ifTrue: 
			[lastPosition := lastPosition + 1.
			character == Character cr 
				ifTrue: 
					[eolPositions add: sourceStream position.
					previousWasCR := true]
				ifFalse: 
					[(previousWasCR not and: [character == Character lf]) 
						ifTrue: [eolPositions add: sourceStream position].
					previousWasCR := false]].
	^character! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ''!
nextPut: anObject 
	^sourceStream nextPut: anObject! !

!SmaCCLineNumberStream methodsFor: 'initialization' stamp: ''!
on: aReadStream 
	sourceStream := aReadStream.
	eolPositions := OrderedCollection with: aReadStream position.
	lastPosition := aReadStream position.
	previousWasCR := false! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ''!
position
	^sourceStream position! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ''!
position: anInteger 
	anInteger > lastPosition 
		ifTrue: 
			[sourceStream position: lastPosition.
			[sourceStream position < anInteger and: [sourceStream atEnd not]] 
				whileTrue: [self next]]
		ifFalse: [sourceStream position: anInteger]! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: ''!
skip: anInteger
	^self position: self position + anInteger! !

!SmaCCLineNumberStream methodsFor: 'accessing' stamp: 'jmb 1/20/2003 21:45'!
upTo: aCharacter 
	| stream char |
	stream := WriteStream on: String new.
	[self atEnd or: [(char := self next) == aCharacter]] 
		whileFalse: [stream nextPut: char].
	^stream contents! !
SmaCCScanner initialize!
